"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis import ShowAs
from zospy.api._ZOSAPI.Analysis.Settings import (
    IAS_,
    IAS_Field,
    IAS_Surface,
    IAS_Wavelength,
)
from zospy.api._ZOSAPI.Analysis.Settings.RMS.RMSField import DataType
from zospy.api._ZOSAPI.Analysis.Settings.RMS.RMSFieldMap import DataType

from . import RMSField, RMSFieldMap

__all__ = (
    "RMSField",
    "RMSFieldMap",
    "FieldDensities",
    "FocusDensities",
    "IAS_RMSField",
    "IAS_RMSFieldMap",
    "IAS_RMSFocus",
    "IAS_RMSLambdaDiagram",
    "Method",
    "Orientations",
    "RayDensities",
    "ReferTo",
    "WaveDensities",
)

class FieldDensities:
    FieldDens_5 = 1
    FieldDens_10 = 2
    FieldDens_15 = 3
    FieldDens_20 = 4
    FieldDens_25 = 5
    FieldDens_30 = 6
    FieldDens_35 = 7
    FieldDens_40 = 8
    FieldDens_45 = 9
    FieldDens_50 = 10
    FieldDens_55 = 11
    FieldDens_60 = 12
    FieldDens_65 = 13
    FieldDens_70 = 14
    FieldDens_75 = 15
    FieldDens_80 = 16
    FieldDens_85 = 17
    FieldDens_90 = 18
    FieldDens_95 = 19
    FieldDens_100 = 20

class FocusDensities:
    FocusDens_5 = 1
    FocusDens_10 = 2
    FocusDens_15 = 3
    FocusDens_20 = 4
    FocusDens_25 = 5
    FocusDens_30 = 6
    FocusDens_35 = 7
    FocusDens_40 = 8
    FocusDens_45 = 9
    FocusDens_50 = 10
    FocusDens_55 = 11
    FocusDens_60 = 12
    FocusDens_65 = 13
    FocusDens_70 = 14
    FocusDens_75 = 15
    FocusDens_80 = 16
    FocusDens_85 = 17
    FocusDens_90 = 18
    FocusDens_95 = 19
    FocusDens_100 = 20

class IAS_RMSField(IAS_):
    @property
    def Data(self) -> DataType: ...
    @property
    def FieldDensity(self) -> FieldDensities: ...
    @property
    def Method(self) -> Method: ...
    @property
    def Orientation(self) -> Orientations: ...
    @property
    def PlotScale(self) -> float: ...
    @property
    def RayDensity(self) -> RayDensities: ...
    @property
    def ReferTo(self) -> ReferTo: ...
    @property
    def RemoveVignettingFactors(self) -> bool: ...
    @property
    def ShowDiffractionLimit(self) -> bool: ...
    @property
    def UseDashes(self) -> bool: ...
    @property
    def UsePolarization(self) -> bool: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @Data.setter
    def Data(self, value: DataType) -> None: ...
    @FieldDensity.setter
    def FieldDensity(self, value: FieldDensities) -> None: ...
    @Method.setter
    def Method(self, value: Method) -> None: ...
    @Orientation.setter
    def Orientation(self, value: Orientations) -> None: ...
    @PlotScale.setter
    def PlotScale(self, value: float) -> None: ...
    @RayDensity.setter
    def RayDensity(self, value: RayDensities) -> None: ...
    @ReferTo.setter
    def ReferTo(self, value: ReferTo) -> None: ...
    @RemoveVignettingFactors.setter
    def RemoveVignettingFactors(self, value: bool) -> None: ...
    @ShowDiffractionLimit.setter
    def ShowDiffractionLimit(self, value: bool) -> None: ...
    @UseDashes.setter
    def UseDashes(self, value: bool) -> None: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...

class IAS_RMSFieldMap(IAS_):
    @property
    def ContourFormat(self) -> str: ...
    @property
    def Data(self) -> DataType: ...
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def MethodType(self) -> Method: ...
    @property
    def PlotScale(self) -> float: ...
    @property
    def RayDensity(self) -> RayDensities: ...
    @property
    def ReferTo(self) -> ReferTo: ...
    @property
    def RemoveVignettingFactors(self) -> bool: ...
    @property
    def ShowAs(self) -> ShowAs: ...
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def UsePolarization(self) -> bool: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def X_FieldSampling(self) -> int: ...
    @property
    def X_FieldSize(self) -> float: ...
    @property
    def Y_FieldSampling(self) -> int: ...
    @property
    def Y_FieldSize(self) -> float: ...
    @ContourFormat.setter
    def ContourFormat(self, value: str) -> None: ...
    @Data.setter
    def Data(self, value: DataType) -> None: ...
    @MethodType.setter
    def MethodType(self, value: Method) -> None: ...
    @PlotScale.setter
    def PlotScale(self, value: float) -> None: ...
    @RayDensity.setter
    def RayDensity(self, value: RayDensities) -> None: ...
    @ReferTo.setter
    def ReferTo(self, value: ReferTo) -> None: ...
    @RemoveVignettingFactors.setter
    def RemoveVignettingFactors(self, value: bool) -> None: ...
    @ShowAs.setter
    def ShowAs(self, value: ShowAs) -> None: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @X_FieldSampling.setter
    def X_FieldSampling(self, value: int) -> None: ...
    @X_FieldSize.setter
    def X_FieldSize(self, value: float) -> None: ...
    @Y_FieldSampling.setter
    def Y_FieldSampling(self, value: int) -> None: ...
    @Y_FieldSize.setter
    def Y_FieldSize(self, value: float) -> None: ...

class IAS_RMSFocus(IAS_):
    @property
    def Data(self) -> DataType: ...
    @property
    def FocusDensity(self) -> FocusDensities: ...
    @property
    def MaximumFocus(self) -> float: ...
    @property
    def Method(self) -> Method: ...
    @property
    def MinimumFocus(self) -> float: ...
    @property
    def PlotScale(self) -> float: ...
    @property
    def RayDensity(self) -> RayDensities: ...
    @property
    def ReferTo(self) -> ReferTo: ...
    @property
    def ShowDiffractionLimit(self) -> bool: ...
    @property
    def UseDashes(self) -> bool: ...
    @property
    def UsePolarization(self) -> bool: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @Data.setter
    def Data(self, value: DataType) -> None: ...
    @FocusDensity.setter
    def FocusDensity(self, value: FocusDensities) -> None: ...
    @MaximumFocus.setter
    def MaximumFocus(self, value: float) -> None: ...
    @Method.setter
    def Method(self, value: Method) -> None: ...
    @MinimumFocus.setter
    def MinimumFocus(self, value: float) -> None: ...
    @PlotScale.setter
    def PlotScale(self, value: float) -> None: ...
    @RayDensity.setter
    def RayDensity(self, value: RayDensities) -> None: ...
    @ReferTo.setter
    def ReferTo(self, value: ReferTo) -> None: ...
    @ShowDiffractionLimit.setter
    def ShowDiffractionLimit(self, value: bool) -> None: ...
    @UseDashes.setter
    def UseDashes(self, value: bool) -> None: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...

class IAS_RMSLambdaDiagram(IAS_):
    @property
    def Data(self) -> DataType: ...
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def Method(self) -> Method: ...
    @property
    def PlotScale(self) -> float: ...
    @property
    def RayDensity(self) -> RayDensities: ...
    @property
    def ReferTo(self) -> ReferTo: ...
    @property
    def ShowDiffractionLimit(self) -> bool: ...
    @property
    def UseDashes(self) -> bool: ...
    @property
    def UsePolarization(self) -> bool: ...
    @property
    def WaveDensity(self) -> WaveDensities: ...
    @Data.setter
    def Data(self, value: DataType) -> None: ...
    @Method.setter
    def Method(self, value: Method) -> None: ...
    @PlotScale.setter
    def PlotScale(self, value: float) -> None: ...
    @RayDensity.setter
    def RayDensity(self, value: RayDensities) -> None: ...
    @ReferTo.setter
    def ReferTo(self, value: ReferTo) -> None: ...
    @ShowDiffractionLimit.setter
    def ShowDiffractionLimit(self, value: bool) -> None: ...
    @UseDashes.setter
    def UseDashes(self, value: bool) -> None: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @WaveDensity.setter
    def WaveDensity(self, value: WaveDensities) -> None: ...

class Method:
    GaussQuad = 0
    RectArray = 1

class Orientations:
    Plus_Y = 0
    Minus_Y = 1
    Plus_X = 2
    Minus_X = 3

class RayDensities:
    RayDens_1 = 1
    RayDens_32x32 = 1
    RayDens_64x64 = 2
    RayDens_2 = 2
    RayDens_128x128 = 3
    RayDens_3 = 3
    RayDens_256x256 = 4
    RayDens_4 = 4
    RayDens_5 = 5
    RayDens_512x512 = 5
    RayDens_1024x1024 = 6
    RayDens_6 = 6
    RayDens_2048x2048 = 7
    RayDens_7 = 7
    RayDens_4096x4096 = 8
    RayDens_8 = 8
    RayDens_8192x8192 = 9
    RayDens_9 = 9
    RayDens_10 = 10
    RayDens_16384x16384 = 10
    RayDens_11 = 11
    RayDens_12 = 12
    RayDens_13 = 13
    RayDens_14 = 14
    RayDens_15 = 15
    RayDens_16 = 16
    RayDens_17 = 17
    RayDens_18 = 18
    RayDens_19 = 19
    RayDens_20 = 20

class ReferTo:
    ChiefRay = 0
    Centroid = 1

class WaveDensities:
    WaveDens_5 = 1
    WaveDens_10 = 2
    WaveDens_15 = 3
    WaveDens_20 = 4
    WaveDens_25 = 5
    WaveDens_30 = 6
    WaveDens_35 = 7
    WaveDens_40 = 8
    WaveDens_45 = 9
    WaveDens_50 = 10
    WaveDens_55 = 11
    WaveDens_60 = 12
    WaveDens_65 = 13
    WaveDens_70 = 14
    WaveDens_75 = 15
    WaveDens_80 = 16
    WaveDens_85 = 17
    WaveDens_90 = 18
    WaveDens_95 = 19
    WaveDens_100 = 20



"""A Python package to communicate with Zemax OpticStudio through its API."""

__version__ = "1.1.1"

import logging

import zospy.api.config as config
from zospy import analyses, functions, solvers
from zospy.api import constants
from zospy.zpcore import ZOS

config.set_decimal_point_and_thousands_separator()

__all__ = (
    "analyses",
    "constants",
    "functions",
    "solvers",
    "ZOS",
)

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = ("EncircledEnergyTypes", "ExtendedSourceTypes", "GeometricLineEdgeSpreadTypes", "ReferToTypes")

class EncircledEnergyTypes(Enum):
    Encircled = 1
    X_Only = 2
    Y_Only = 3
    Ensquared = 4

class ExtendedSourceTypes(Enum):
    Encircled = 1
    X_Only = 2
    Y_Only = 3
    Ensquared = 4
    X_Distrib = 5
    Y_Distrib = 6

class GeometricLineEdgeSpreadTypes(Enum):
    LineEdge = 1
    Line = 2
    Edge = 3

class ReferToTypes(Enum):
    ChiefRay = 0
    Centroid = 1
    Vertex = 2



import logging
import os
import sys
import winreg

import clr  # noqa

import zospy.api.constants
from zospy.utils import clrutils

logger = logging.getLogger(__name__)


def get_zos_root():
    """Obtains the Zemax OpticStudio folder from the windows registry.

    Returns
    -------
    filepath: str
        The root folder of the installed Zemax Version

    Raises
    ------
    FileNotFoundError
        If either Zemax OpticStudio cannot be found in the Windows Registry or the Zemax OpticStudio root folder could
        not be obtained from the registry.
    """
    logger.info("Obtaining Zemax Location from Windows Registry")

    # Search for Zemax OpticStudio in the Windows Registry
    try:
        regkey = winreg.OpenKey(winreg.HKEY_CURRENT_USER, r"Software\Zemax")
    except FileNotFoundError as e:
        logger.exception("Zemax cannot be found in the registry, " "the following error is raised:\n{}".format(e))
        raise e

    # Obtain the Zemax OpticStudio root path from the registry
    try:
        zemaxfolder, _ = winreg.QueryValueEx(regkey, "ZemaxRoot")
        logger.info("Found ZemaxFolder at {}".format(zemaxfolder))
    except FileNotFoundError as e:
        logger.exception(
            "The Zemax Root folder cannot be found in the registry key, " "the following error is raised:\n{}".format(e)
        )
        raise e
    finally:  # Clean up
        logger.debug("Closing registry key")
        winreg.CloseKey(regkey)
        logger.debug("Registry key closed")

    return zemaxfolder


def load_zosapi_nethelper(filepath=None, preload=False):
    """Loads the ZOSAPI_NetHelper.

    Parameters
    ----------
    filepath: str, optional
        The location of the ZOSAPI_NetHelper. If not supplied, the location will be deducted from the windows registry.
    preload: bool
        Specifies if CLR should import with preload on (True) or off (False). Defaults to False).

    Returns
    -------
    ZOSAPI_NetHelper: netModuleObject
        The ZOSAPI_NetHelper as module object
    """
    if not filepath:  # get the path through the windows registry
        zos_root = get_zos_root()
        znh_filepath = os.path.join(zos_root, r"ZOS-API\Libraries\ZOSAPI_NetHelper.dll")
    else:
        znh_filepath = filepath

    logger.debug("Adding reference {} to clr".format(znh_filepath))
    sys.path.append(os.path.basename(znh_filepath))
    clr.AddReference(znh_filepath)

    logger.debug("Importing ZOSAPI_NetHelper")
    znh = __import__("ZOSAPI_NetHelper", globals(), locals(), [], 0)
    logger.info("ZOSAPI_NetHelper imported successfully")

    return znh


def load_zosapi(zosapi_nethelper=None, zemaxdirectory=None, preload=False):
    """Load the ZOS-API using the ZOSAPI NetHelper DLL.

    Parameters
    ----------
    zosapi_nethelper: netModuleObject, optional
        The loaded ZOSAPI_NetHelper DLL. If None, the zemaxdirectory parameter will be used. Note that either the
        zosapi_nethelper or the zemaxdirectory has to be specified, not both.
    zemaxdirectory: str, optional
        The directory containing the Zemax DLLs. If None, the directory will be acquired through the zosapi_nethelper.
        Note that either the zosapi_nethelper or the zemaxdirectory has to be specified, not both.
    preload: bool
        Specifies if CLR should import with preload on (True) or off (False). Defaults to False).

    Returns
    -------
    zosapi: netModuleObject
        The ZOSAPI module.

    Raises
    ------
    ValueError:
        If either both or None of zosapi_nethelper and zemaxdirectory are specified.
    FileNotFoundError:
        If one of ZOSAPI_Interfaces.dll or ZOSAPI.dll cannot be found.
    """
    if not any((zosapi_nethelper, zemaxdirectory)):
        raise ValueError("Either the zosapi_nethelper or the zemaxdirectory should be specified.")
    elif all((zosapi_nethelper, zemaxdirectory)):
        raise ValueError("Only one of zosapi_nethelper and zemaxdirectory should be specified.")
    else:
        pass

    # Get the Zemax OpticStudio directory and add it to the path
    if zosapi_nethelper:
        logger.info("Obtaining Zemax Directory from ZOSAPI_NetHelper")
        zosapi_nethelper.ZOSAPI_Initializer.Initialize()
        zos_dir = zosapi_nethelper.ZOSAPI_Initializer.GetZemaxDirectory()
        logger.info("Zemax OpticStudio found at {}".format(zos_dir))
    else:
        logger.info("Zemax Directory specified by user ({})".format(zemaxdirectory))
        zos_dir = zemaxdirectory
    sys.path.append(zos_dir)

    logger.info("Searching and registering ZOSAPI DLLs")
    for dll in ["ZOSAPI_Interfaces", "ZOSAPI"]:
        if clr.FindAssembly(dll):
            logger.debug("{}.dll found".format(dll))
            clr.AddReference(dll)
            logger.info("{} imported to clr".format(dll))
        else:
            logger.critical("Cannot locate {}.dll".format(dll))
            raise FileNotFoundError("Cannot locate {}.dll in {}".format(dll, zos_dir))

    logger.debug("Checking content of ZOSAPI_Interfaces.dll")
    content = clrutils.reflect_dll_content(os.path.join(zos_dir, "ZOSAPI_Interfaces.dll"))

    logger.debug("Loading ZOSAPI")
    zosapi = __import__("ZOSAPI", globals(), locals(), [], 0)
    logger.debug("ZOSAPI loaded")

    logger.debug("Loading nested namespaces")
    for nsp in content["namespaces"]:
        if nsp == "ZOSAPI":
            continue
        __import__(nsp, globals(), locals(), [], 0)
        logger.debug("Nested namespace {} preloaded".format(nsp))

    zospy.api.constants._construct_from_zosapi_and_enumkeys(zosapi, content["enums"])  # noqa

    return zosapi



"""Zemax OpticStudio analyses from the PSF category."""

from __future__ import annotations

import pandas as pd

from zospy import utils
from zospy.analyses.base import AnalysisResult, AttrDict, OnComplete, new_analysis
from zospy.api import constants


def huygens_psf(
    oss,
    pupil_sampling: str | int = "32x32",
    image_sampling: str | int = "32x32",
    image_delta: float = 0,
    rotation: float = 0,
    wavelength: str | int = "All",
    field: str | int = 1,
    psftype: constants.Analysis.Settings.HuygensPsfTypes | str = "Linear",
    show_as: constants.Analysis.HuygensShowAsTypes | str = "Surface",
    use_polarization: bool = False,
    use_centroid: bool = False,
    normalize: bool = False,
    oncomplete: OnComplete | str = OnComplete.Close,
) -> AnalysisResult:
    """Wrapper around the OpticStudio Huygens PSF.

    Parameters
    ----------
    oss: zospy.core.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    pupil_sampling: str | int
        The pupil sampling, either string (e.g. '64x64') or int. The integer will be treated as a ZOSAPI Constants
        integer.
    image_sampling: str | int
        The image sampling, either string (e.g. '64x64') or int. The integer will be treated as a ZOSAPI Constants
        integer.
    image_delta: float | int
        The image delta
    rotation: int
        The rotation, should be one of [0, 90, 180, 270].
    wavelength: str | int
        The wavelength number that is to be used. Either 'All' or an integer specifying the wavelength number.
        Defaults to 'All'.
    field: str | int
        The field number that is to be used. Either 'All' or an integer specifying the field number. Defaults to 1.
    psftype: str | int
        The PSF type (e.g. 'Linear') that is calculated. Defaults to 'Linear'.
    show_as: str | int
        Defines how the data is showed within OpticStudio. Defaults to 'Surface'
    use_polarization: bool
        Defines if polarization is used. Defaults to False.
    use_centroid: bool
        Defines if centroid is used. Defaults to False.
    normalize: bool
        Defines if normalization is used. Defaults to False.
    oncomplete: OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to 'Close'.

    Returns
    -------
    AnalysisResult
        A HuygensPsf analysis result
    """  # ToDo check if default for filed is correct
    analysistype = constants.Analysis.AnalysisIDM.HuygensPsf

    # Create analysis
    analysis = new_analysis(oss, analysistype)

    # Apply settings
    analysis.Settings.PupilSampleSize = getattr(
        constants.Analysis.SampleSizes, utils.zputils.standardize_sampling(pupil_sampling)
    )
    analysis.Settings.ImageSampleSize = getattr(
        constants.Analysis.SampleSizes, utils.zputils.standardize_sampling(image_sampling)
    )
    analysis.Settings.ImageDelta = image_delta
    analysis.Settings.Rotation = getattr(constants.Analysis.Settings.Rotations, f"Rotate_{rotation}")
    analysis.set_wavelength(wavelength)
    analysis.set_field(field)
    analysis.Settings.Type = constants.process_constant(constants.Analysis.Settings.HuygensPsfTypes, psftype)
    analysis.Settings.ShowAsType = constants.process_constant(constants.Analysis.HuygensShowAsTypes, show_as)
    analysis.Settings.UsePolarization = use_polarization
    analysis.Settings.UseCentroid = use_centroid
    analysis.Settings.Normalize = normalize

    # Calculate
    analysis.ApplyAndWaitForCompletion()

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Get settings
    settings = pd.Series(name="Settings", dtype=object)

    settings.loc["PupilSampleSize"] = str(analysis.Settings.PupilSampleSize)
    settings.loc["ImageSampleSize"] = str(analysis.Settings.ImageSampleSize)
    settings.loc["ImageDelta"] = analysis.Settings.ImageDelta
    settings.loc["Rotation"] = int(str(analysis.Settings.Rotation).split("_")[1])
    settings.loc["Wavelength"] = analysis.get_wavelength()
    settings.loc["Field"] = analysis.get_field()
    settings.loc["Type"] = str(analysis.Settings.Type)
    settings.loc["ShowAsType"] = str(analysis.Settings.ShowAsType)
    settings.loc["UsePolarization"] = analysis.Settings.UsePolarization
    settings.loc["UseCentroid"] = analysis.Settings.UseCentroid
    settings.loc["Normalize"] = analysis.Settings.Normalize

    # Get data
    if analysis.Results.NumberOfDataGrids <= 0:
        data = None
    elif analysis.Results.NumberOfDataGrids == 1:
        data = utils.zputils.unpack_datagrid(analysis.Results.DataGrids[0])
    else:
        data = AttrDict()
        for ii in range(analysis.Results.NumberOfDataGrids):
            desc = analysis.Results.DataGrids[ii].Description
            key = desc if desc != "" else str(ii)
            data[key] = utils.zputils.unpack_datagrid(analysis.Results.DataGrids[ii])

    result = AnalysisResult(
        analysistype=str(analysistype),
        data=data,
        settings=settings,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Editors.LDE import SurfaceColumn
from zospy.api._ZOSAPI.Editors.NCE import ObjectColumn
from zospy.api._ZOSAPI.SystemData import FieldColumn

from . import LDE, MCE, MFE, NCE, TDE

__all__ = (
    "LDE",
    "MCE",
    "MFE",
    "NCE",
    "TDE",
    "CellDataType",
    "EditorType",
    "IEditor",
    "IEditor",
    "IEditorCell",
    "IEditorCell",
    "IEditorRow",
    "IEditorRow",
    "ISolveAplanatic",
    "ISolveAutomatic",
    "ISolveCenterOfCurvature",
    "ISolveChiefRayAngle",
    "ISolveChiefRayHeight",
    "ISolveChiefRayNormal",
    "ISolveCocentricRadius",
    "ISolveCocentricSurface",
    "ISolveCompensator",
    "ISolveConfigPickup",
    "ISolveData",
    "ISolveData",
    "ISolveDuplicateSag",
    "ISolveDuplicateSag",
    "ISolveEdgeThickness",
    "ISolveElementPower",
    "ISolveFieldPickup",
    "ISolveFixed",
    "ISolveFNumber",
    "ISolveInvertSag",
    "ISolveInvertSag",
    "ISolveMarginalRayAngle",
    "ISolveMarginalRayHeight",
    "ISolveMarginalRayNormal",
    "ISolveMaterialModel",
    "ISolveMaterialOffset",
    "ISolveMaterialSubstitute",
    "ISolveMaximum",
    "ISolveNone",
    "ISolveObjectPickup",
    "ISolveOpticalPathDifference",
    "ISolvePickupChiefRay",
    "ISolvePosition",
    "ISolvePupilPosition",
    "ISolveSurfacePickup",
    "ISolveThermalPickup",
    "ISolveVariable",
    "ISolveZPLMacro",
    "ReflectTransmitCode",
    "SampleSides",
    "Samplings",
    "SolveStatus",
    "SolveType",
)

class CellDataType:
    Integer = 0
    Double = 1
    String = 2

class EditorType:
    LDE = 0
    NCE = 1
    MFE = 2
    TDE = 3
    MCE = 4

class IEditor:
    def AddRow(self) -> IEditorRow: ...
    def DeleteAllRows(self) -> int: ...
    def DeleteRowAt(self, pos: int) -> bool: ...
    def DeleteRowsAt(self, pos: int, numberOfRows: int) -> int: ...
    @property
    def Editor(self) -> EditorType: ...
    @property
    def MaxColumn(self) -> int: ...
    @property
    def MinColumn(self) -> int: ...
    @property
    def NumberOfRows(self) -> int: ...
    def GetRowAt(self, pos: int) -> IEditorRow: ...
    def HideEditor(self) -> None: ...
    def InsertRowAt(self, pos: int) -> IEditorRow: ...
    def ShowEditor(self) -> bool: ...

class IEditorCell:
    def CreateSolveType(self, type: SolveType) -> ISolveData: ...
    def FillAvailableSolveTypes(self, Length: int) -> tuple[list[SolveType]]: ...
    @property
    def Col(self) -> int: ...
    @property
    def DataType(self) -> CellDataType: ...
    @property
    def DoubleValue(self) -> float: ...
    @property
    def Header(self) -> str: ...
    @property
    def IntegerValue(self) -> int: ...
    @property
    def IsActive(self) -> bool: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def Row(self) -> IEditorRow: ...
    @property
    def Solve(self) -> SolveType: ...
    @property
    def Value(self) -> str: ...
    def GetAvailableSolveTypes(self) -> list[SolveType]: ...
    def GetNumberOfSolveTypes(self) -> int: ...
    def GetSolveData(self) -> ISolveData: ...
    def IsSolveTypeSupported(self, st: SolveType) -> bool: ...
    def MakeSolveFixed(self) -> bool: ...
    def MakeSolveVariable(self) -> bool: ...
    @DoubleValue.setter
    def DoubleValue(self, value: float) -> None: ...
    @IntegerValue.setter
    def IntegerValue(self, value: int) -> None: ...
    @Value.setter
    def Value(self, value: str) -> None: ...
    def SetSolveData(self, Data: ISolveData) -> SolveStatus: ...

class IEditorRow:
    @property
    def Bookmark(self) -> str: ...
    @property
    def Editor(self) -> IEditor: ...
    @property
    def IsValidRow(self) -> bool: ...
    @property
    def RowIndex(self) -> int: ...
    @property
    def RowTypeName(self) -> str: ...
    def GetCellAt(self, pos: int) -> IEditorCell: ...
    @Bookmark.setter
    def Bookmark(self, value: str) -> None: ...

class ISolveAplanatic(ISolveData):
    pass

class ISolveAutomatic(ISolveData):
    pass

class ISolveCenterOfCurvature(ISolveData):
    @property
    def RefSurface(self) -> int: ...
    @RefSurface.setter
    def RefSurface(self, value: int) -> None: ...

class ISolveChiefRayAngle(ISolveData):
    @property
    def Angle(self) -> float: ...
    @Angle.setter
    def Angle(self, value: float) -> None: ...

class ISolveChiefRayHeight(ISolveData):
    @property
    def Height(self) -> float: ...
    @Height.setter
    def Height(self, value: float) -> None: ...

class ISolveChiefRayNormal(ISolveData):
    pass

class ISolveCocentricRadius(ISolveData):
    @property
    def WithSurface(self) -> int: ...
    @WithSurface.setter
    def WithSurface(self, value: int) -> None: ...

class ISolveCocentricSurface(ISolveData):
    @property
    def AboutSurface(self) -> int: ...
    @AboutSurface.setter
    def AboutSurface(self, value: int) -> None: ...

class ISolveCompensator(ISolveData):
    @property
    def RefSurface(self) -> int: ...
    @property
    def Sum(self) -> float: ...
    @RefSurface.setter
    def RefSurface(self, value: int) -> None: ...
    @Sum.setter
    def Sum(self, value: float) -> None: ...

class ISolveConfigPickup(ISolveData):
    @property
    def Configuration(self) -> int: ...
    @property
    def Offset(self) -> float: ...
    @property
    def Operand(self) -> int: ...
    @property
    def ScaleFactor(self) -> float: ...
    @property
    def SupportsOffset(self) -> bool: ...
    @property
    def SupportsScale(self) -> bool: ...
    @Configuration.setter
    def Configuration(self, value: int) -> None: ...
    @Offset.setter
    def Offset(self, value: float) -> None: ...
    @Operand.setter
    def Operand(self, value: int) -> None: ...
    @ScaleFactor.setter
    def ScaleFactor(self, value: float) -> None: ...

class ISolveData:
    @property
    def _S_Aplanatic(self) -> ISolveAplanatic: ...
    @property
    def _S_Automatic(self) -> ISolveAutomatic: ...
    @property
    def _S_CenterOfCurvature(self) -> ISolveCenterOfCurvature: ...
    @property
    def _S_ChiefRayAngle(self) -> ISolveChiefRayAngle: ...
    @property
    def _S_ChiefRayHeight(self) -> ISolveChiefRayHeight: ...
    @property
    def _S_ChiefRayNormal(self) -> ISolveChiefRayNormal: ...
    @property
    def _S_CocentricRadius(self) -> ISolveCocentricRadius: ...
    @property
    def _S_CocentricSurface(self) -> ISolveCocentricSurface: ...
    @property
    def _S_Compensator(self) -> ISolveCompensator: ...
    @property
    def _S_ConfigPickup(self) -> ISolveConfigPickup: ...
    @property
    def _S_DuplicateSag(self) -> ISolveDuplicateSag: ...
    @property
    def _S_EdgeThickness(self) -> ISolveEdgeThickness: ...
    @property
    def _S_ElementPower(self) -> ISolveElementPower: ...
    @property
    def _S_FieldPickup(self) -> ISolveFieldPickup: ...
    @property
    def _S_Fixed(self) -> ISolveFixed: ...
    @property
    def _S_FNumber(self) -> ISolveFNumber: ...
    @property
    def _S_InvertSag(self) -> ISolveInvertSag: ...
    @property
    def _S_MarginalRayAngle(self) -> ISolveMarginalRayAngle: ...
    @property
    def _S_MarginalRayHeight(self) -> ISolveMarginalRayHeight: ...
    @property
    def _S_MarginalRayNormal(self) -> ISolveMarginalRayNormal: ...
    @property
    def _S_MaterialModel(self) -> ISolveMaterialModel: ...
    @property
    def _S_MaterialOffset(self) -> ISolveMaterialOffset: ...
    @property
    def _S_MaterialSubstitute(self) -> ISolveMaterialSubstitute: ...
    @property
    def _S_Maximum(self) -> ISolveMaximum: ...
    @property
    def _S_None(self) -> ISolveNone: ...
    @property
    def _S_ObjectPickup(self) -> ISolveObjectPickup: ...
    @property
    def _S_OpticalPathDifference(self) -> ISolveOpticalPathDifference: ...
    @property
    def _S_PickupChiefRay(self) -> ISolvePickupChiefRay: ...
    @property
    def _S_Position(self) -> ISolvePosition: ...
    @property
    def _S_PupilPosition(self) -> ISolvePupilPosition: ...
    @property
    def _S_SurfacePickup(self) -> ISolveSurfacePickup: ...
    @property
    def _S_ThermalPickup(self) -> ISolveThermalPickup: ...
    @property
    def _S_Variable(self) -> ISolveVariable: ...
    @property
    def _S_ZPLMacro(self) -> ISolveZPLMacro: ...
    @property
    def IsValid(self) -> bool: ...
    @property
    def Type(self) -> SolveType: ...

class ISolveDuplicateSag:
    @property
    def Surface(self) -> int: ...
    @Surface.setter
    def Surface(self, value: int) -> None: ...

class ISolveEdgeThickness(ISolveData):
    @property
    def RadialHeight(self) -> float: ...
    @property
    def Thickness(self) -> float: ...
    @RadialHeight.setter
    def RadialHeight(self, value: float) -> None: ...
    @Thickness.setter
    def Thickness(self, value: float) -> None: ...

class ISolveElementPower(ISolveData):
    @property
    def Power(self) -> float: ...
    @Power.setter
    def Power(self, value: float) -> None: ...

class ISolveFieldPickup(ISolveData):
    @property
    def Column(self) -> FieldColumn: ...
    @property
    def Field(self) -> int: ...
    @property
    def Offset(self) -> float: ...
    @property
    def ScaleFactor(self) -> float: ...
    def IsPickupFromCurrentColumn(self) -> bool: ...
    def MakePickupFromCurrentColumn(self) -> None: ...
    @Column.setter
    def Column(self, value: FieldColumn) -> None: ...
    @Field.setter
    def Field(self, value: int) -> None: ...
    @Offset.setter
    def Offset(self, value: float) -> None: ...
    @ScaleFactor.setter
    def ScaleFactor(self, value: float) -> None: ...

class ISolveFixed(ISolveData):
    pass

class ISolveFNumber(ISolveData):
    @property
    def FNumber(self) -> float: ...
    @FNumber.setter
    def FNumber(self, value: float) -> None: ...

class ISolveInvertSag:
    @property
    def Surface(self) -> int: ...
    @Surface.setter
    def Surface(self, value: int) -> None: ...

class ISolveMarginalRayAngle(ISolveData):
    @property
    def Angle(self) -> float: ...
    @Angle.setter
    def Angle(self, value: float) -> None: ...

class ISolveMarginalRayHeight(ISolveData):
    @property
    def Height(self) -> float: ...
    @property
    def PupilZone(self) -> float: ...
    @Height.setter
    def Height(self, value: float) -> None: ...
    @PupilZone.setter
    def PupilZone(self, value: float) -> None: ...

class ISolveMarginalRayNormal(ISolveData):
    pass

class ISolveMaterialModel(ISolveData):
    @property
    def AbbeVd(self) -> float: ...
    @property
    def dPgF(self) -> float: ...
    @property
    def IndexNd(self) -> float: ...
    @property
    def VaryAbbe(self) -> bool: ...
    @property
    def VarydPgF(self) -> bool: ...
    @property
    def VaryIndex(self) -> bool: ...
    @AbbeVd.setter
    def AbbeVd(self, value: float) -> None: ...
    @dPgF.setter
    def dPgF(self, value: float) -> None: ...
    @IndexNd.setter
    def IndexNd(self, value: float) -> None: ...
    @VaryAbbe.setter
    def VaryAbbe(self, value: bool) -> None: ...
    @VarydPgF.setter
    def VarydPgF(self, value: bool) -> None: ...
    @VaryIndex.setter
    def VaryIndex(self, value: bool) -> None: ...

class ISolveMaterialOffset(ISolveData):
    @property
    def NdOffset(self) -> float: ...
    @property
    def VdOffset(self) -> float: ...
    @NdOffset.setter
    def NdOffset(self, value: float) -> None: ...
    @VdOffset.setter
    def VdOffset(self, value: float) -> None: ...

class ISolveMaterialSubstitute(ISolveData):
    @property
    def Catalog(self) -> str: ...
    @Catalog.setter
    def Catalog(self, value: str) -> None: ...

class ISolveMaximum(ISolveData):
    pass

class ISolveNone(ISolveData):
    pass

class ISolveObjectPickup(ISolveData):
    @property
    def Column(self) -> ObjectColumn: ...
    @property
    def Object(self) -> int: ...
    @property
    def Offset(self) -> float: ...
    @property
    def ScaleFactor(self) -> float: ...
    @property
    def SupportsOffset(self) -> bool: ...
    @property
    def SupportsScale(self) -> bool: ...
    def IsPickupFromCurrentColumn(self) -> bool: ...
    def MakePickupFromCurrentColumn(self) -> None: ...
    @Column.setter
    def Column(self, value: ObjectColumn) -> None: ...
    @Object.setter
    def Object(self, value: int) -> None: ...
    @Offset.setter
    def Offset(self, value: float) -> None: ...
    @ScaleFactor.setter
    def ScaleFactor(self, value: float) -> None: ...

class ISolveOpticalPathDifference(ISolveData):
    @property
    def OPD(self) -> float: ...
    @property
    def PupilZone(self) -> float: ...
    @OPD.setter
    def OPD(self, value: float) -> None: ...
    @PupilZone.setter
    def PupilZone(self, value: float) -> None: ...

class ISolvePickupChiefRay(ISolveData):
    @property
    def Field(self) -> int: ...
    @property
    def Wavelength(self) -> int: ...
    @Field.setter
    def Field(self, value: int) -> None: ...
    @Wavelength.setter
    def Wavelength(self, value: int) -> None: ...

class ISolvePosition(ISolveData):
    @property
    def FromSurface(self) -> int: ...
    @property
    def Length(self) -> float: ...
    @FromSurface.setter
    def FromSurface(self, value: int) -> None: ...
    @Length.setter
    def Length(self, value: float) -> None: ...

class ISolvePupilPosition(ISolveData):
    pass

class ISolveSurfacePickup(ISolveData):
    @property
    def Column(self) -> SurfaceColumn: ...
    @property
    def Offset(self) -> float: ...
    @property
    def ScaleFactor(self) -> float: ...
    @property
    def SupportsOffset(self) -> bool: ...
    @property
    def SupportsScale(self) -> bool: ...
    @property
    def Surface(self) -> int: ...
    def IsPickupFromCurrentColumn(self) -> bool: ...
    def MakePickupFromCurrentColumn(self) -> None: ...
    @Column.setter
    def Column(self, value: SurfaceColumn) -> None: ...
    @Offset.setter
    def Offset(self, value: float) -> None: ...
    @ScaleFactor.setter
    def ScaleFactor(self, value: float) -> None: ...
    @Surface.setter
    def Surface(self, value: int) -> None: ...

class ISolveThermalPickup(ISolveData):
    @property
    def Configuration(self) -> int: ...
    @Configuration.setter
    def Configuration(self, value: int) -> None: ...

class ISolveVariable(ISolveData):
    pass

class ISolveZPLMacro(ISolveData):
    @property
    def Macro(self) -> str: ...
    def GetAvailableMacros(self) -> list[str]: ...
    @Macro.setter
    def Macro(self, value: str) -> None: ...

class ReflectTransmitCode:
    Success = 0
    NoReflectDataInFile = 1
    NoTransmitDataInFile = 2

class SampleSides:
    Front = 0
    Back = 1

class Samplings:
    FiveDegrees = 0
    TwoDegrees = 1
    OneDegree = 2

class SolveStatus:
    Success = 0
    InvalidSolveType = 1
    InvalidRow = 2
    InvalidColumn = 3
    PostSurfaceStopOnly = 4
    InvalidMacro = 5
    Failed = 10000

class SolveType:
    # None = 0
    Fixed = 1
    Variable = 2
    SurfacePickup = 3
    ZPLMacro = 4
    MarginalRayAngle = 5
    MarginalRayHeight = 6
    ChiefRayAngle = 7
    MarginalRayNormal = 8
    ChiefRayNormal = 9
    Aplanatic = 10
    ElementPower = 11
    CocentricSurface = 12
    ConcentricSurface = 12
    CocentricRadius = 13
    ConcentricRadius = 13
    FNumber = 14
    ChiefRayHeight = 15
    EdgeThickness = 16
    OpticalPathDifference = 17
    Position = 18
    Compensator = 19
    CenterOfCurvature = 20
    PupilPosition = 21
    MaterialSubstitute = 22
    MaterialOffset = 23
    MaterialModel = 24
    Automatic = 25
    Maximum = 26
    PickupChiefRay = 27
    ObjectPickup = 28
    ConfigPickup = 29
    ThermalPickup = 30
    MarginPercent = 31
    CA_fill = 32
    DIA_fill = 33
    DuplicateSag = 34
    InvertSag = 35
    FieldPickup = 10000



"""Constants for the ZOS-API.

Submodule used for package wide access to all ZOS api constants. Note that the constant-naming within this module breaks
pep, but is kept as such to be in-sync with the api documentation. Constants should in general be accessed through
'zospy.constants' (or 'zp.constants'). All constants are obtained dynamically from the api. Therefore, they are only
available after running the code stated under examples.

Examples
--------
import zospy as zp
zos = zp.ZOS()
zos.wakeup()
zp.constants
"""
from __future__ import annotations

import itertools as _itertools
import logging as _logging
from types import SimpleNamespace as _SimpleNamespace
from typing import Iterable, TypeVar

from zospy.api._ZOSAPI_constants import *  # noqa
from zospy.utils import clrutils as _clrutils
from zospy.utils import pyutils as _pyutils

_logger = _logging.getLogger(__name__)


# Note: the above imports (and all but one function) are named private, to incease clarity on which Module components
# are constants and which are not.


def _itertools_joinfunc(*args):
    return ".".join(args)


def _construct_from_zosapi_and_enumkeys(zosapi, zosapi_enumkeys):
    """Constructs the constants from the zosapi and zosapi_enumkeys.

    The acquired constants are added to the zp.constants namespace

    Parameters
    ----------
    zosapi:
        A ZOSAPI instance.
    zosapi_enumkeys: list[str]
        The enumeration keys obtained from can be obtained through zp.utils.clrutils.reflect_dll_content()

    Returns
    -------
    None
    """
    zosapi_enumkeys = sorted(zosapi_enumkeys)
    added_namespaces = set()
    for enumkey in zosapi_enumkeys:
        subkeys = enumkey.split(".")

        assert len(subkeys) > 1  # should at least be 2

        if len(subkeys) == 2:  # No nesting
            clrattr = getattr(zosapi, subkeys[-1], None)

            # Set constant
            globals()[subkeys[-1]] = _clrutils.system_enum_to_namedtuple(clrattr)

        else:  # with nesting
            base = subkeys[1]
            nsp_parts = list(_itertools.accumulate(subkeys[2:-1], func=_itertools_joinfunc))

            if base not in added_namespaces:
                globals()[base] = _SimpleNamespace()  # Create the base as simplenamespace
                added_namespaces.add(base)
            for nsp in nsp_parts:
                if ".".join((base, nsp)) in added_namespaces:  # check if already added
                    continue
                else:
                    _pyutils.rsetattr(globals()[base], nsp, _SimpleNamespace())  # add nested objects
                    added_namespaces.add(".".join((base, nsp)))

            clrattr = _pyutils.rgetattr(zosapi, ".".join(subkeys[1:]), None)

            # set constants
            _pyutils.rsetattr(globals()[base], ".".join(subkeys[2:]), _clrutils.system_enum_to_namedtuple(clrattr))


Constant = TypeVar("Constant")


def process_constant(constant: Iterable[Constant], value: Constant | str | None) -> Constant:
    if (value is None or value == "None") and hasattr(constant, "None_"):
        return getattr(constant, "None_")
    elif isinstance(value, str) and hasattr(constant, value):
        return getattr(constant, value)
    elif value in constant:
        return value

    raise ValueError(f"Constant {type(constant).__name__} does not contain value {str(value)}")


def get_constantname_by_value(constant_tuple, value):
    """Obtain a constant name from a value.

    Parameters
    ----------
    constant_tuple: tuple
        The set of constants used to look up the value
    value: int
        The value for which the constant name is to be found

    Returns
    -------
    str
        The constant name
    """
    try:
        if isinstance(value, int):
            return constant_tuple._fields[value]  # noqa

        return constant_tuple._fields[constant_tuple.index(value)]  # noqa
    except KeyError:
        raise ValueError(f"None of the constants has value {value} assigned")



"""Zemax OpticStudio analyses from the Rays and Spots category."""

from __future__ import annotations

import os
import re
from io import StringIO
from tempfile import mkstemp
from typing import Any

import numpy as np
import pandas as pd

import zospy.api.config as _config
from zospy.analyses.base import AnalysisResult, AttrDict, OnComplete, new_analysis
from zospy.api import constants
from zospy.zpcore import OpticStudioSystem


def _structure_ray_trace_data_result(line_list: list[str]) -> dict[str, Any]:
    """Structures the result of a ray trace data report.

    Parameters
    ----------
    line_list: list[str]
        The line list obtained by reading in the results

    Returns
    -------
    dict
        The results structured in a dictionary
    """
    # Create output dict
    res = AttrDict()

    # Add header
    hdr = line_list[0].strip().replace("\ufeff", "")
    res["Header"] = hdr

    # Register landmarks
    landmarks = [
        "Real Ray Trace Data",
        "Paraxial Ray Trace Data",
        "Trace of Paraxial Y marginal, U marginal, Y chief, U chief only",
    ]
    comp = re.compile(rf'^({"|".join(landmarks)})', re.IGNORECASE)
    inds = {
        comp.match(line).group().lower(): {"rnum": rnum, "name": comp.match(line).group()}
        for rnum, line in enumerate(line_list)
        if comp.match(line)
    }

    order_of_appearance = [item[0] for item in sorted(inds.items(), key=lambda x: x[1]["rnum"])]

    # Add general lens data
    for target in inds.keys():
        section_start = inds[target]["rnum"] + 1  # + 1 to ignore section title

        # Get endpoint
        if order_of_appearance.index(target) == len(inds) - 1:
            section_end = len(line_list)
        else:
            target2 = order_of_appearance[order_of_appearance.index(target) + 1]
            section_end = inds[target2]["rnum"]

        # Read as dataframe
        df = pd.read_csv(
            StringIO("".join(line_list[section_start:section_end]).replace(" ", "")),
            delimiter="\t",
            decimal=_config.DECIMAL_POINT,
        )

        # Recover case of target
        keyname = inds[target]["name"].replace(" ", "")

        # Add to result dictionary
        res[keyname] = df

    return res


def single_ray_trace(
    oss: OpticStudioSystem,
    hx: float = 0,
    hy: float = 0,
    px: float = 0,
    py: float = 1,
    wavelength: str | int = 1,
    field: str | int = 0,
    raytrace_type: constants.Analysis.Settings.Aberrations.RayTraceType | str = "DirectionCosines",
    global_coordinates: bool = False,
    oncomplete: OnComplete | str = OnComplete.Close,
    txtoutfile: str | None = None,
) -> AnalysisResult:
    """Wrapper around the OpticStudio Single Ray Trace Analysis.

    Due to limitations in the ZOS-API, the output is obtained by writing the OpticStudio results to a file and
    subsequently reading in this file

    Parameters
    ----------
    oss: zospy.core.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    hx: float
        Normalized X Field Coord. Defaults to 0.
    hy: float
        Normalized Y Field Coord. Defaults to 0.
    px: float
        Normalized X Pupil Coord. Defaults to 0.
    py: float
        Normalized Y Pupil Coord. Defaults to 1.
    raytrace_type: zospy.api.constants.Analysis.Settings.Aberrations.RayTraceType
        The Ray Trace type (e.g. 'DirectionCosines') that is calculated. Defaults to 'DirectionCosines'.
    wavelength: int
        The wavelength number that is to be used. Must be an integer specifying the wavelength number.
        Defaults to 1.
    field: int
        The field number that is to be used. Must be an integer specifying the field number. "Arbitrary" is 0. Defaults
        to 0.
    global_coordinates: bool
        Defines if global coordinates are used. Defaults to False.
    oncomplete: OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to 'Close'.
    txtoutfile: str | None
        The textfile to which the OpticStudio output is saved. If None, a temporary file will be created and deleted. If
        string, it should be a full system path with '.txt' as extension, after which the file will be saved in that
        location. Defaults to None.

    Returns
    -------
    AnalysisResult
        A Single Ray Trace result. Next to the standard data, the raw text return obtained from the analysis
        will be present under 'RawTextData', and the txtoutfile under 'TxtOutFile'.
    """
    analysis_type = constants.Analysis.AnalysisIDM.RayTrace

    # Temp file
    if txtoutfile is None:
        fd, txtoutfile = mkstemp(suffix=".txt", prefix="zospy_")
        os.close(fd)
        cleantxt = True
    else:
        if not txtoutfile.endswith(".txt"):
            raise ValueError('txtfile should end with ".txt"')
        cleantxt = False

    analysis = new_analysis(oss, analysis_type)

    analysis.Settings.Hx = hx
    analysis.Settings.Hy = hy
    analysis.Settings.Px = px
    analysis.Settings.Py = py
    analysis.set_wavelength(wavelength)
    analysis.set_field(field)
    analysis.Settings.Type = constants.process_constant(
        constants.Analysis.Settings.Aberrations.RayTraceType, raytrace_type
    )
    analysis.Settings.UseGlobal = global_coordinates

    # Run analysis
    analysis.ApplyAndWaitForCompletion()

    # Get results
    analysis.Results.GetTextFile(txtoutfile)
    line_list = [line for line in open(txtoutfile, "r", encoding=oss._ZOS.get_txtfile_encoding())]
    data = _structure_ray_trace_data_result(line_list)

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Get settings
    settings = pd.Series(name="Settings", dtype=object)

    settings.loc["Hx"] = analysis.Settings.Hx
    settings.loc["Hy"] = analysis.Settings.Hy
    settings.loc["Px"] = analysis.Settings.Px
    settings.loc["Py"] = analysis.Settings.Py
    settings.loc["Wavelength"] = analysis.get_wavelength()
    settings.loc["Field"] = analysis.get_field()
    settings.loc["Type"] = str(analysis.Settings.Type)
    settings.loc["GlobalCoordinates"] = analysis.Settings.UseGlobal

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=data,
        settings=settings,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
        RawTextData=line_list,
        TxtOutFile=txtoutfile,
    )  # Set additional params

    # cleanup if needed
    if cleantxt:
        os.remove(txtoutfile)

    return analysis.complete(oncomplete, result)


def ray_fan(
    oss: OpticStudioSystem,
    plot_scale: float = 0,
    number_of_rays: int = 20,
    field: str | int = "All",
    wavelength: str | int = "All",
    tangential: constants.Analysis.Settings.Fans.TangentialAberrationComponent | str = "Aberration_Y",
    sagittal: constants.Analysis.Settings.Fans.SagittalAberrationComponent | str = "Aberration_X",
    surface: str | int = "Image",
    use_dashes: bool = False,
    vignetted_pupil: bool = True,
    check_apertures: bool = True,
    oncomplete: OnComplete | str = OnComplete.Close,
) -> AnalysisResult:
    """Wrapper around the OpticStudio Ray Fan Analysis.

    Parameters
    ----------
    oss: zospy.core.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    plot_scale: int | float
        Sets the maximum vertical scale for the plots. When 0, automatic scaling is used. Defaults to 0.
    number_of_rays: int
        This is the number of rays traced on each side of the origin of the plot.
    field: str or int
        The field number that is to be used. Either 'All' or an integer specifying the field number. Defaults to 'All'.
    wavelength: str | int
        The wavelength number that is to be used. Either 'All' or an integer specifying the wavelength number.
        Defaults to 'All'.
    tangential: str | int
        The aberration component that is plotted for the tangential fan. Accepts sting ('Aberration_Y' or
        'Aberration_X') or int (respectively 0 and 1). Defaults to 'Aberration_Y'.
    sagittal: str | int
        The aberration component that is plotted for the sagittal fan. Accepts sting ('Aberration_X' or
        'Aberration_Y') or int (respectively 0 and 1). Defaults to 'Aberration_X'.
    surface: str | int
        The surface that is to be analyzed. Either 'Image', 'Object' or an integer. Defaults to 'Image'.
    use_dashes: bool
        Defines whether solid lines or dashes are used to differentiate curves. Defaults to False.
    vignetted_pupil: bool
        Defines whether the pupil axis is scaled to the unvignetted pupil or not. Defaults to True.
    check_apertures: bool
        Defines whether only rays that pass all surface apertures are drawn or not. Defaults to True.
    oncomplete: OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to 'Close'.

    Returns
    -------
    AnalysisResult
        A Ray Fan result.
    """
    # Set up analysis type
    analysis_type = constants.Analysis.AnalysisIDM.RayFan
    analysis = new_analysis(oss, analysis_type)

    # Settings for ray fan
    analysis.set_field(field)
    analysis.set_surface(surface)
    analysis.set_wavelength(wavelength)
    analysis.Settings.NumberOfRays = number_of_rays
    analysis.Settings.PlotScale = plot_scale
    analysis.Settings.CheckApertures = check_apertures
    analysis.Settings.VignettedPupil = vignetted_pupil
    analysis.Settings.UseDashes = use_dashes
    analysis.Settings.Sagittal = constants.process_constant(
        constants.Analysis.Settings.Fans.SagittalAberrationComponent, sagittal
    )
    analysis.Settings.Tangential = constants.process_constant(
        constants.Analysis.Settings.Fans.TangentialAberrationComponent, tangential
    )

    # Run analysis
    analysis.ApplyAndWaitForCompletion()

    # Parse results
    results = analysis.GetResults()
    data = AttrDict()
    data["Header"] = "Ray Fan Data"
    for ii in range(results.DataSeries.Length):
        # get raw .NET data into numpy array
        ds = results.GetDataSeries(ii)
        xRaw = np.asarray(tuple(ds.XData.Data))
        yRaw = np.asarray(tuple(ds.YData.Data))

        # Reshape data
        x = xRaw
        y = yRaw.reshape(ds.YData.Data.GetLength(0), ds.YData.Data.GetLength(1))

        # Make data frame
        df = {ds.XLabel: x}
        for jj, label in enumerate(ds.SeriesLabels):
            df[label] = y[:, jj]
        df = pd.DataFrame(data=df)

        # Add data to dictionary
        data[ds.Description] = df

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Get settings
    settings = pd.Series(name="Settings", dtype=object)
    settings.loc["Field"] = analysis.get_field()
    settings.loc["Surface"] = analysis.Settings.Surface.GetSurfaceNumber()
    settings.loc["Wavelength"] = analysis.get_wavelength()
    settings.loc["NumberOfRays"] = analysis.Settings.NumberOfRays
    settings.loc["PlotScale"] = analysis.Settings.PlotScale
    settings.loc["CheckApertures"] = analysis.Settings.CheckApertures
    settings.loc["VignettedPupil"] = analysis.Settings.VignettedPupil
    settings.loc["UseDashes"] = analysis.Settings.UseDashes
    settings.loc["Sagittal"] = str(analysis.Settings.Sagittal)
    settings.loc["Tangential"] = str(analysis.Settings.Tangential)

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=data,
        settings=settings,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )  # Set additional params

    return analysis.complete(oncomplete, result)



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

from . import General, LMx, Optimization, RayTrace, Tolerancing

__all__ = (
    "General",
    "LMx",
    "Optimization",
    "RayTrace",
    "Tolerancing",
    "CriticalRayType",
    "HPCEnvironments",
    "HPCNodeSize",
    "HPCRunState",
    "MaterialFormulas",
    "MaterialStatuses",
    "RayPatternOption",
    "RunStatus",
    "VertexOrder",
)

class CriticalRayType(Enum):
    Chief = 0
    Marginal = 1
    Grid = 2
    Ring = 3
    Y_Fan = 4
    X_Fan = 5
    XY_Fan = 6
    List = 7

class HPCEnvironments(Enum):
    OnPremise = 0
    AWSKubernetes = 1
    AzureKubernetes = 2

class HPCNodeSize(Enum):
    Default = 0
    Tiny = 1
    Small = 2
    Medium = 3
    Large = 4
    XLarge = 5

class HPCRunState(Enum):
    NotRunning = 0
    Initializing = 1
    ClusterAllocating = 2
    UploadingData = 3
    Queued = 4
    RunStarting = 5
    WaitingForResults = 6
    Complete = 7

class MaterialFormulas(Enum):
    Schott = 1
    Sellmeier1 = 2
    Herzberger = 3
    Sellmeier2 = 4
    Conrady = 5
    Sellmeier3 = 6
    Handbook1 = 7
    Handbook2 = 8
    Sellmeier4 = 9
    Extended = 10
    Sellmeier5 = 11
    Extended2 = 12
    Extended3 = 13

class MaterialStatuses(Enum):
    Standard = 0
    Preferred = 1
    Obsolete = 2
    Special = 3
    Melt = 4

class RayPatternOption(Enum):
    XyFan = 0
    XFan = 1
    YFan = 2
    ChiefAndRing = 3
    List = 4
    Grid = 6
    ChiefAndMarginals = 8

class RunStatus(Enum):
    Completed = 0
    FailedToStart = 1
    TimedOut = 2
    InvalidTimeout = 3

class VertexOrder(Enum):
    First = 0
    Second = 1
    Third = 2



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis.Settings import IAS_, IAS_Field, IAS_Wavelength
from zospy.api._ZOSAPI.Analysis.Tolerancing.QuickYield import IAS_QYField
from zospy.api._ZOSAPI.Tools.Tolerancing import MonteCarloStatistics

from . import QuickYield

__all__ = (
    "QuickYield",
    "IAS_QuickYield",
    "IAS_TolerancingHistogram",
    "IAS_TolerancingYield",
    "ITolerancingOperand",
    "ITolerancingOperand",
    "ITolerancingOperands",
    "ITolerancingOperands",
    "QYCompensations",
    "QYCompensatorStrategy",
    "QYPrecisions",
)

class IAS_QuickYield(IAS_):
    @property
    def Compensation(self) -> QYCompensations: ...
    @property
    def CompensatorStrategy(self) -> QYCompensatorStrategy: ...
    @property
    def Configuration(self) -> int: ...
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def NumMonteCarlo(self) -> int: ...
    @property
    def Precision(self) -> QYPrecisions: ...
    @property
    def PupilSampling(self) -> int: ...
    @property
    def QYField(self) -> IAS_QYField: ...
    @property
    def Statistic(self) -> MonteCarloStatistics: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @Compensation.setter
    def Compensation(self, value: QYCompensations) -> None: ...
    @CompensatorStrategy.setter
    def CompensatorStrategy(self, value: QYCompensatorStrategy) -> None: ...
    @Configuration.setter
    def Configuration(self, value: int) -> None: ...
    @NumMonteCarlo.setter
    def NumMonteCarlo(self, value: int) -> None: ...
    @Precision.setter
    def Precision(self, value: QYPrecisions) -> None: ...
    @PupilSampling.setter
    def PupilSampling(self, value: int) -> None: ...
    @Statistic.setter
    def Statistic(self, value: MonteCarloStatistics) -> None: ...

class IAS_TolerancingHistogram(IAS_):
    @property
    def Filename(self) -> str: ...
    @property
    def MaxValue(self) -> float: ...
    @property
    def MinValue(self) -> float: ...
    @property
    def NumBins(self) -> int: ...
    @property
    def Operand(self) -> int: ...
    def GetOperands(self) -> ITolerancingOperands: ...
    def GetToleranceFiles(self) -> list[str]: ...
    @Filename.setter
    def Filename(self, value: str) -> None: ...
    @MaxValue.setter
    def MaxValue(self, value: float) -> None: ...
    @MinValue.setter
    def MinValue(self, value: float) -> None: ...
    @NumBins.setter
    def NumBins(self, value: int) -> None: ...
    @Operand.setter
    def Operand(self, value: int) -> None: ...
    def UseSystemTolerances(self) -> None: ...

class IAS_TolerancingYield(IAS_):
    @property
    def Filename(self) -> str: ...
    @property
    def Operand(self) -> int: ...
    def GetOperands(self) -> ITolerancingOperands: ...
    def GetToleranceFiles(self) -> list[str]: ...
    @Filename.setter
    def Filename(self, value: str) -> None: ...
    @Operand.setter
    def Operand(self, value: int) -> None: ...
    def UseSystemTolerances(self) -> None: ...

class ITolerancingOperand:
    @property
    def OperandName(self) -> str: ...
    @property
    def ParameterData(self) -> str: ...

class ITolerancingOperands:
    @property
    def NumberOfOperands(self) -> int: ...
    @property
    def Operands(self) -> list[ITolerancingOperand]: ...
    def GetOperand(self, operandIndex: int) -> ITolerancingOperand: ...

class QYCompensations:
    Standard = 0
    High = 1
    VeryHigh = 2

class QYCompensatorStrategy:
    OptimizeAllDampedLeastSquares = 0
    ParaxialFocus = 1
    Ignore = 2
    OptimizeAllOrthogonalDescent = 3

class QYPrecisions:
    Standard = 0
    High = 1
    VeryHigh = 2



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = ("DateTimeType", "EncodingType", "LanguageType", "ShowLineAsType")

class DateTimeType(Enum):
    None_ = 0
    Date = 1
    DateTime = 2

class EncodingType(Enum):
    ANSI = 0
    Unicode = 1

class LanguageType(Enum):
    Chinese = 0
    English = 1
    Japanese = 4

class ShowLineAsType(Enum):
    TextAbove = 0
    FileName = 1
    ConfigurationNumber = 2



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis.Settings import (
    IAS_,
    IAS_Field,
    IAS_Surface,
    IAS_Wavelength,
)

__all__ = ("IAS_Fan", "SagittalAberrationComponent", "TangentialAberrationComponent")

class IAS_Fan(IAS_):
    @property
    def CheckApertures(self) -> bool: ...
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def NumberOfRays(self) -> int: ...
    @property
    def PlotScale(self) -> float: ...
    @property
    def Sagittal(self) -> SagittalAberrationComponent: ...
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Tangential(self) -> TangentialAberrationComponent: ...
    @property
    def UseDashes(self) -> bool: ...
    @property
    def VignettedPupil(self) -> bool: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @CheckApertures.setter
    def CheckApertures(self, value: bool) -> None: ...
    @NumberOfRays.setter
    def NumberOfRays(self, value: int) -> None: ...
    @PlotScale.setter
    def PlotScale(self, value: float) -> None: ...
    @Sagittal.setter
    def Sagittal(self, value: SagittalAberrationComponent) -> None: ...
    @Tangential.setter
    def Tangential(self, value: TangentialAberrationComponent) -> None: ...
    @UseDashes.setter
    def UseDashes(self, value: bool) -> None: ...
    @VignettedPupil.setter
    def VignettedPupil(self, value: bool) -> None: ...

class SagittalAberrationComponent:
    Aberration_X = 0
    Aberration_Y = 1

class TangentialAberrationComponent:
    Aberration_Y = 0
    Aberration_X = 1



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = ("ColorRaysBy", "Patterns", "Reference", "ShowScales")

class ColorRaysBy(Enum):
    Fields = 0
    Waves = 1
    Config = 2
    Wavelength = 3

class Patterns(Enum):
    Square = 0
    Hexapolar = 1
    Dithered = 2

class Reference(Enum):
    ChiefRay = 0
    Centroid = 1
    Middle = 2
    Vertex = 3

class ShowScales(Enum):
    ScaleBar = 0
    Box = 1
    Cross = 2
    Circle = 3



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = ("HuygensMtfTypes", "MtfDataTypes", "MtfTypes", "ShowAsTypes", "SurfaceMtfTypes")

class HuygensMtfTypes(Enum):
    Modulation = 0

class MtfDataTypes(Enum):
    Average = 0
    Tangential = 1
    Sagittal = 2

class MtfTypes(Enum):
    Modulation = 0
    Real = 1
    Imaginary = 2
    Phase = 3
    SquareWave = 4

class ShowAsTypes(Enum):
    GreyScale = 0
    GreyScaleInverted = 1
    FalseColor = 2
    FalseColorInverted = 3

class SurfaceMtfTypes(Enum):
    MTF_Incoherent = 0
    Real_Incoherent = 1
    Imaginary_Incoherent = 2
    MTF_Coherent = 3
    Real_Coherent = 4
    Imaginary_Coherent = 5



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

from . import Analysis, Common, Editors, Preferences, SystemData, Tools, Wizards

__all__ = (
    "Analysis",
    "Common",
    "Editors",
    "Preferences",
    "SystemData",
    "Tools",
    "Wizards",
    "LensUpdateMode",
    "LicenseStatusType",
    "SessionModes",
    "STARUpdateMode",
    "SystemType",
    "UpdateStatus",
    "ZOSAPI_Mode",
)

class LensUpdateMode(Enum):
    None_ = 0
    EditorsOnly = 1
    AllWindows = 2

class LicenseStatusType(Enum):
    Unknown = 0
    KeyNotWorking = 1
    NewLicenseNeeded = 2
    StandardEdition = 3
    ProfessionalEdition = 4
    PremiumEdition = 5
    TooManyInstances = 6
    NotAuthorized = 7
    KeyNotFound = 8
    KeyExpired = 9
    Timeout = 10
    InstanceConflict = 11
    OpticsViewer = 12
    OpticStudioHPCEdition = 13
    EnterpriseEdition = 14

class SessionModes(Enum):
    FromPreferences = 0
    SessionOn = 1
    SessionOff = 2

class STARUpdateMode(Enum):
    Normal = 0
    Suspended = 1

class SystemType(Enum):
    Sequential = 0
    NonSequential = 1

class UpdateStatus(Enum):
    NotChecked = 0
    UpToDate = 1
    AvailableEligible = 2
    AvailableIneligible = 3
    CheckFailed = -2
    NotSupported = -1

class ZOSAPI_Mode(Enum):
    Server = 0
    Operand = 1
    Plugin = 2
    UserAnalysis = 3
    UserAnalysisSettings = 4



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Common import ZemaxColor
from zospy.api._ZOSAPI.Editors import IEditor, IEditorCell, IEditorRow

__all__ = ("IMCERow", "IMultiConfigEditor", "MultiConfigOperandType")

class IMCERow(IEditorRow):
    def AvailableConfigOperandTypes(self) -> list[MultiConfigOperandType]: ...
    def ChangeType(self, type: MultiConfigOperandType) -> bool: ...
    @property
    def IsActive(self) -> bool: ...
    @property
    def OperandNumber(self) -> int: ...
    @property
    def Param1(self) -> int: ...
    @property
    def Param1Enabled(self) -> bool: ...
    @property
    def Param2(self) -> int: ...
    @property
    def Param2Enabled(self) -> bool: ...
    @property
    def Param3(self) -> int: ...
    @property
    def Param3Enabled(self) -> bool: ...
    @property
    def RowColor(self) -> ZemaxColor: ...
    @property
    def Type(self) -> MultiConfigOperandType: ...
    @property
    def TypeName(self) -> str: ...
    def GetOperandCell(self, configuration: int) -> IEditorCell: ...
    @Param1.setter
    def Param1(self, value: int) -> None: ...
    @Param2.setter
    def Param2(self, value: int) -> None: ...
    @Param3.setter
    def Param3(self, value: int) -> None: ...
    @RowColor.setter
    def RowColor(self, value: ZemaxColor) -> None: ...

class IMultiConfigEditor(IEditor):
    def AddConfiguration(self, withPickups: bool) -> bool: ...
    def AddOperand(self) -> IMCERow: ...
    def CopyOperands(self, fromOperandNumber: int, NumberOfOperands: int, toOperandNumber: int) -> int: ...
    def CopyOperandsFrom(
        self, fromEditor: IMultiConfigEditor, fromOperandNumber: int, NumberOfOperands: int, toOperandNumber: int
    ) -> int: ...
    def DeleteAllConfigurations(self) -> bool: ...
    def DeleteConfiguration(self, ConfigurationNumber: int) -> bool: ...
    @property
    def CurrentConfiguration(self) -> int: ...
    @property
    def FirstConfiguration(self) -> int: ...
    @property
    def LastConfiguration(self) -> int: ...
    @property
    def NumberOfConfigurations(self) -> int: ...
    @property
    def NumberOfOperands(self) -> int: ...
    @property
    def RowToOperandOffset(self) -> int: ...
    def GetOperandAt(self, OperandNumber: int) -> IMCERow: ...
    def HideMCE(self) -> None: ...
    def InsertConfiguration(self, ConfigurationNumber: int, withPickups: bool) -> bool: ...
    def InsertNewOperandAt(self, OperandNumber: int) -> IMCERow: ...
    def MakeSingleConfiguration(self) -> None: ...
    def MakeSingleConfigurationOpt(self, deleteMFEOperands: bool) -> None: ...
    def NextConfiguration(self) -> bool: ...
    def PrevConfiguration(self) -> bool: ...
    def RemoveOperandAt(self, OperandNumber: int) -> bool: ...
    def RemoveOperandsAt(self, OperandNumber: int, numOperands: int) -> int: ...
    def RunTool_MakeThermal(
        self,
        existingConfigurationIndex: int,
        numberOfThemalConfigs: int,
        minTemp: float,
        maxTemp: float,
        sortBySurface: bool,
    ) -> None: ...
    def SetCurrentConfiguration(self, ConfigurationNumber: int) -> bool: ...
    def ShowMCE(self) -> bool: ...

class MultiConfigOperandType:
    MOFF = 0
    AICN = 1
    AFOC = 2
    APDF = 3
    APDT = 4
    APDX = 5
    APDY = 6
    APER = 7
    APMN = 8
    APMX = 9
    APTP = 10
    CADX = 11
    CADY = 12
    CAOR = 13
    CATX = 14
    CATY = 15
    CATZ = 16
    CBDX = 17
    CBDY = 18
    CBOR = 19
    CBTX = 20
    CBTY = 21
    CBTZ = 22
    CONN = 23
    COTN = 24
    CPCN = 25
    CROR = 26
    CRSR = 27
    CRVT = 28
    CSP1 = 29
    CSP2 = 30
    CWGT = 31
    EDVA = 32
    FLTP = 33
    FLWT = 34
    FVAN = 35
    FVCX = 36
    FVCY = 37
    FVDX = 38
    FVDY = 39
    GCRS = 40
    GLSS = 41
    GPEX = 42
    GPEY = 43
    GPIU = 44
    GPJX = 45
    GPJY = 46
    GPPX = 47
    GPPY = 48
    GQPO = 49
    HOLD = 50
    IGNM = 51
    IGNR = 52
    LTTL = 53
    MABB = 54
    MCOM = 55
    MDPG = 56
    MIND = 57
    MTFU = 58
    NCOM = 59
    NCOT = 60
    NGLS = 61
    NPAR = 62
    NPOS = 63
    NPRO = 64
    PAR1 = 65
    PAR2 = 66
    PAR3 = 67
    PAR4 = 68
    PAR5 = 69
    PAR6 = 70
    PAR7 = 71
    PAR8 = 72
    PRAM = 73
    PRES = 74
    PRWV = 75
    PSCX = 76
    PSCY = 77
    PSHX = 78
    PSHY = 79
    PSHZ = 80
    PSP1 = 81
    PSP2 = 82
    PSP3 = 83
    PUCN = 84
    PXAR = 85
    RAAM = 86
    SATP = 87
    SDIA = 88
    SDRW = 89
    STPS = 90
    SWCN = 91
    TCEX = 92
    TELE = 93
    TEMP = 94
    THIC = 95
    TSP1 = 96
    TSP2 = 97
    TSP3 = 98
    UDAF = 99
    WAVE = 100
    WLWT = 101
    XFIE = 102
    YFIE = 103
    OPDR = 104
    SRTS = 105
    MCHI = 106
    IGTO = 107
    FTAN = 108
    FCMM = 109
    CHZN = 110
    MCSD = 111



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = ("DataType", "FoucaultShowAs", "KnifeType", "Types")

class DataType(Enum):
    Computed = 0
    Reference = 1
    Difference = 2

class FoucaultShowAs(Enum):
    Surface = 0
    Contour = 1
    GreyScale = 2
    InverseGreyScale = 3
    FalseColor = 4
    InverseFalseColor = 5
    X_CrossSection = 6
    Y_CrossSection = 7

class KnifeType(Enum):
    Horiz_Above = 0
    Horiz_Below = 1
    Vert_Right = 2
    Vert_Left = 3

class Types(Enum):
    Linear = 0
    Log_Minus_3 = 1
    Log_Minus_6 = 2
    Log_Minus_9 = 3
    Log_Minus_12 = 4
    Log_Minus_15 = 5
    Log_Minus_18 = 6



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = ("FftPsfType", "PsfRotation", "PsfSampling")

class FftPsfType(Enum):
    Linear = 0
    Log = 1
    Phase = 2
    Real = 3
    Imaginary = 4

class PsfRotation(Enum):
    CW0 = 0
    CW90 = 1
    CW180 = 2
    CW270 = 3

class PsfSampling(Enum):
    PsfS_32x32 = 1
    PsfS_64x64 = 2
    PsfS_128x128 = 3
    PsfS_256x256 = 4
    PsfS_512x512 = 5
    PsfS_1024x1024 = 6
    PsfS_2048x2048 = 7
    PsfS_4096x4096 = 8
    PsfS_8192x8192 = 9
    PsfS_16384x16384 = 10



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = ("SettingsDataType", "ZemaxColor", "ZemaxOpacity")

class SettingsDataType(Enum):
    None_ = 0
    Integer = 1
    Double = 2
    Float = 3
    String = 4
    Byte = 5
    Boolean = 6

class ZemaxColor(Enum):
    Default = 0
    Color1 = 1
    Color2 = 2
    Color3 = 3
    Color4 = 4
    Color5 = 5
    Color6 = 6
    Color7 = 7
    Color8 = 8
    Color9 = 9
    Color10 = 10
    Color11 = 11
    Color12 = 12
    Color13 = 13
    Color14 = 14
    Color15 = 15
    Color16 = 16
    Color17 = 17
    Color18 = 18
    Color19 = 19
    Color20 = 20
    Color21 = 21
    Color22 = 22
    Color23 = 23
    Color24 = 24

class ZemaxOpacity(Enum):
    P100 = 0
    P90 = 1
    P80 = 2
    P70 = 3
    P60 = 4
    P50 = 5
    P40 = 6
    P30 = 7
    P20 = 8
    P10 = 9
    P00 = 10



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = ("ToleranceColumn", "ToleranceOperandType")

class ToleranceColumn(Enum):
    Param1 = 2
    Param2 = 3
    Param3 = 4
    Nominal = 5
    Min = 6
    Max = 7
    Comment = 8

class ToleranceOperandType(Enum):
    TOFF = 0
    TRAD = 1
    TCUR = 2
    TFRN = 3
    TTHI = 4
    TCON = 5
    TSDI = 6
    TSDX = 7
    TSDY = 8
    TSDR = 9
    TSTX = 10
    TSTY = 11
    TIRX = 12
    TIRY = 13
    TIRR = 14
    TPAR = 15
    TEDV = 16
    TEDX = 17
    TEDY = 18
    TEDR = 19
    TETX = 20
    TETY = 21
    TETZ = 22
    TUDX = 23
    TUDY = 24
    TUTX = 25
    TUTY = 26
    TUTZ = 27
    TIND = 28
    TABB = 29
    TEXI = 30
    SAVE = 31
    COMP = 32
    CPAR = 33
    CEDV = 34
    CMCO = 35
    STAT = 36
    TWAV = 37
    TNPS = 38
    TNPA = 39
    TMCO = 40
    TEZI = 41
    TCMU = 42
    SEED = 43
    COMM = 44
    TPAI = 45
    TCIO = 46
    TCEO = 47
    TNMA = 48
    CNPA = 49
    CNPS = 50
    TRLX = 51
    TRLY = 52
    TRLR = 53
    TARX = 54
    TARY = 55
    TARR = 56



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = (
    "CriterionComps",
    "CriterionFields",
    "Criterions",
    "MonteCarloStatistics",
    "QSCriterions",
    "SetupCaches",
    "SetupChanges",
    "SetupModes",
    "SetupPolynomials",
    "TolerancingColumnName",
    "TolerancingOperand",
    "TolerancingParameterName",
)

class CriterionComps(Enum):
    OptimizeAll_DLS = 0
    NONSEQOptimizeAll = 0
    ParaxialFocus = 1
    NONSEQNone = 1
    NONSEQOptimizeAll_OD = 2
    None_ = 2
    OptimizeAll_OD = 3

class CriterionFields(Enum):
    NONSEQUserDefined = 0
    Y_Symmetric = 0
    XY_Symmetric = 1
    UserDefined = 2

class Criterions(Enum):
    NONSEQMeritFunction = 0
    RMSSpotRadius = 0
    RMSSpotX = 1
    NONSEQUserScript = 1
    RMSSpotY = 2
    RMSWavefront = 3
    MeritFunction = 4
    GeometricMTFAverage = 5
    GeometricMTFTan = 6
    GeometricMTFSag = 7
    DiffMTFAverage = 8
    DiffMTFTan = 9
    DiffMTFSag = 10
    BoresightError = 11
    RMSAngularRadius = 12
    RMSAngularX = 13
    RMSAngularY = 14
    UserScript = 15

class MonteCarloStatistics(Enum):
    Normal = 0
    Uniform = 1
    Parabolic = 2

class QSCriterions(Enum):
    RMSSpotRadius = 0
    RMSSpotX = 1
    RMSSpotY = 2
    RMSWavefront = 3
    BoresightError = 4
    RMSAngularRadius = 5
    RMSAngularX = 6
    RMSAngularY = 7

class SetupCaches(Enum):
    RecomputeAll = 0
    RecomputeChanged = 1
    UsePolynomial = 2

class SetupChanges(Enum):
    LinearDifference = 0
    RSSDifference = 1

class SetupModes(Enum):
    Sensitivity = 0
    InverseLimit = 1
    InverseIncrement = 2
    SkipSensitivity = 3

class SetupPolynomials(Enum):
    None_ = 0
    ThreeMinusTerm = 1
    FiveMinusTerm = 2

class TolerancingColumnName(Enum):
    UserMeritFunction = 11959
    UserScript = 11960
    RmsSpotRadius = 11961
    RmsSpotX = 11962
    RmsSpotY = 11963
    RmsWavefrontError = 11964
    MtfGeometricAverage = 11965
    MtfGeometricTangential = 11966
    MtfGeometricSaggital = 11967
    MtfDiffractionAverage = 11968
    MtfDiffractionTangential = 11969
    MtfDiffractionSagittal = 11970
    BoresightError = 11971
    RmsAngularRadius = 11972
    RmsAngularX = 11973
    RmsAngularY = 11974
    BackFocusChange = 75417
    Unknown = -1

class TolerancingOperand(Enum):
    NULL = 0
    TRAD = 1
    TCUR = 2
    TFRN = 3
    TTHI = 4
    TCON = 5
    TSDI = 6
    TSDX = 7
    TSDY = 8
    TSTX = 9
    TSTY = 10
    TIRX = 11
    TIRY = 12
    TIRR = 13
    TPAR = 14
    TEDV = 15
    TEDX = 16
    TEDY = 17
    TETX = 18
    TETY = 19
    TETZ = 20
    TUDX = 21
    TUDY = 22
    TUTX = 23
    TUTY = 24
    TUTZ = 25
    TIND = 26
    TABB = 27
    TEXI = 28
    SAVE = 29
    COMP = 30
    CPAR = 31
    CEDV = 32
    CMCO = 33
    STAT = 34
    TWAV = 35
    TNPS = 36
    TNPA = 37
    TMCO = 38
    TEZI = 39
    TCMU = 40
    SEED = 41
    COMM = 42
    TPAI = 43
    TCIO = 44
    TCEO = 45
    TNMA = 46
    CNPA = 47
    CNPS = 48
    TRLX = 49
    TRLY = 50
    TRLR = 51
    TARX = 52
    TARY = 53
    TARR = 54
    TEDR = 55
    TSDR = 56
    Unknown = -1

class TolerancingParameterName(Enum):
    CompensatorSetting = 11400
    ConfigurationNumber = 16304
    CommentField = 20004
    NominalValue = 20566
    Surface = 20567
    MinimumValue = 20568
    MaximumValue = 20569
    Surface1 = 20570
    Surface2 = 20571
    AdjustSurface = 20572
    ParameterNumber = 20573
    Layer = 20574
    MaximumNumber = 20576
    MinimumNumber = 20577
    Code = 20578
    RowNumber = 20579
    ConfigurationParameter = 20580
    Object = 20585
    Data = 20586
    FieldNumber = 20683
    SamplingSetting = 20694
    FieldSetting = 50027
    TdeRowNumber = 75418
    Undefined = -1



from __future__ import annotations

from zospy.api import _ZOSAPI, constants


def object_change_type(obj: _ZOSAPI.Editors.NCE.INCERow, new_type: constants.Editors.NCE.ObjectType | str):
    """Simple function to change the object type in the NCE.

    Parameters
    ----------
    obj: ZOSAPI.Editors.NCE.INCERow
        The Row/Object for which the change is to be made.
    new_type: zospy.constants.Editors.NCE.ObjectType | str
        The new object type, either string (e.g. 'StandardLens') or int. The integer will be treated as if obtained from
        zp.constants.Editors.NCE.ObjectType.

    Returns
    -------
    None

    Examples
    --------
    >>> import zospy as zp
    >>> zos = zp.ZOS()
    >>> zos.connect_as_extension()
    >>> oss = zos.get_primary_system()
    >>> oss.make_nonsequential()
    >>> newobj = oss.NCE.InsertNewObjectAt(1)
    >>> zp.functions.nce.object_change_type(newobj, zp.constants.Editors.NCE.ObjectType.StandardLens)
    """
    new_type = constants.process_constant(constants.Editors.NCE.ObjectType, new_type)

    # Apply
    new_surface_type_settings = obj.GetObjectTypeSettings(new_type)
    obj.ChangeType(new_surface_type_settings)


def find_object_by_comment(
    nce: _ZOSAPI.Editors.NCE, comment: str, case_sensitive: bool = True
) -> list[_ZOSAPI.Editors.NCE.INCERow]:
    """Returns a list of objects from the NCE that have the supplied string as Comment.

    In case of multiple matches, the objects are returned in ascending order.

    Parameters
    ----------
    nce: ZOSAPI.Editors.NCE
        The Non-sequential Component Editor (NCE).
    comment: str
        String that is searched for in the Comment column of the NCE.
    case_sensitive: bool=False
        Flag that specifies whether the search is case-sensitive (default value) or not.

    Returns
    -------
    list[ZOSAPI.Editors.NCE.INCERow]
        A list of object in the NCE that have a Comment column value which matches the comment argument.

    Examples
    --------
    >>> import zospy as zp
    >>> zos = zp.ZOS()
    >>> zos.connect_as_extension()
    >>> oss = zos.get_primary_system()
    >>> oss.make_nonsequential()
    >>> newobj1 = oss.NCE.InsertNewObjectAt(1)
    >>> newobj1.Comment = 'aa'
    >>> newobj2 = oss.NCE.InsertNewObjectAt(1)
    >>> newobj2.Comment = 'bb'
    >>> newobj3 = oss.NCE.InsertNewObjectAt(1)
    >>> newobj3.Comment = 'aA'
    >>> find_object_by_comment(oss.NCE, 'aa')
    """
    # Number of objects in the NCE
    number_of_objects = nce.NumberOfObjects

    # Is the search case-sensitive?
    if not case_sensitive:
        # If the search is NOT case-sensitive put comment argument in all small letters
        comment = comment.lower()

    # Initialize list of objects with matching comment
    matching_objects = []

    # Loop over the objects in the NCE and check if their comment matches
    for object_index in range(1, number_of_objects + 1):
        # Retrieve current object
        current_object = nce.GetObjectAt(object_index)

        # Retrieve current object comment
        object_comment = current_object.Comment

        # Is the search case-sensitive?
        if not case_sensitive:
            # If the search is NOT case-sensitive put current object comment in all small letters
            object_comment = object_comment.lower()

        # If the comment matches, append the corresponding object to the return list
        if comment == object_comment:
            matching_objects.append(current_object)

    # Return list of objects with matching comment
    return matching_objects


def get_object_data(obj: _ZOSAPI.Editors.NCE.INCERow) -> _ZOSAPI.Editors.NCE.IObject:
    """Returns the object-specific data.

    Parameters
    ----------
    obj: ZOSAPI.Editors.NCE.INCERow
        The Row/Object for which the data is requested.

    Returns
    -------
    ZOSAPI.Editors.NCE.IObject
        The object-specific data with the inherited implementation.

    Examples
    --------
    >>> import zospy as zp
    >>> zos = zp.ZOS()
    >>> zos.connect_as_extension()
    >>> oss = zos.get_primary_system()
    >>> oss.make_nonsequential()
    >>> detector_object = oss.NCE.GetObjectAt(1)
    >>> detector_type = detector_object.GetObjectTypeSettings(zp.constants.Editors.NCE.ObjectType.DetectorRectangle)
    >>> detector_object.ChangeType(detector_type)
    >>> detector_data = zp.functions.nce.get_object_data(detector_object)
    >>> number_of_x_pixels = detector_data.NumberXPixels
    """
    return obj.ObjectData.__implementation__



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = (
    "CriterionTypes",
    "DefaultAndDegrees",
    "DefaultAndFringes",
    "OptimizationTypes",
    "PupilArmsCount",
    "ReferenceTypes",
    "ToleranceGrade",
    "ToleranceVendor",
    "WizardType",
)

class CriterionTypes(Enum):
    Wavefront = 0
    Contrast = 1
    Spot = 2
    Angular = 3

class DefaultAndDegrees(Enum):
    Default = 0
    Degrees = 1

class DefaultAndFringes(Enum):
    Default = 0
    Fringes = 1
    Percent = 2

class OptimizationTypes(Enum):
    RMS = 0
    PTV = 1

class PupilArmsCount(Enum):
    Arms_6 = 0
    Arms_8 = 1
    Arms_10 = 2
    Arms_12 = 3

class ReferenceTypes(Enum):
    Centroid = 0
    ChiefRay = 1
    Unreferenced = 2

class ToleranceGrade(Enum):
    Commercial = 0
    Precision = 1
    HighPrecision = 2
    CellPhoneLens = 3

class ToleranceVendor(Enum):
    Asphericon = 0
    EdmundOptics = 1
    Generic = 2
    LaCroix = 3
    Optimax = 4

class WizardType(Enum):
    NSCOptimization = 0
    NSCBitmap = 1
    NSCRoadwayLighting = 2
    SEQOptimization = 3
    NSCTolerance = 4
    SEQTolerance = 5



"""Zemax OpticStudio System Viewers."""

from __future__ import annotations

from zospy.analyses.base import AnalysisResult, OnComplete, new_analysis
from zospy.api import constants
from zospy.zpcore import OpticStudioSystem


def cross_section(
    oss: OpticStudioSystem,
    oncomplete: OnComplete | str = OnComplete.Release,
) -> AnalysisResult:
    """Wrapper around the OpticStudio Cross-Section viewer.

    Parameters
    ----------
    oss: zospy.core.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    oncomplete: OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to
        'Release'.

    Returns
    -------
    AnalysisResult
        A Cross-Section analysis result. As the viewers do not return data, the AnalysisResult can mainly be used to
        further control the analysis when oncomplete is set to `Sustain`.
    """
    analysis_type = constants.Analysis.AnalysisIDM.Draw2D

    analysis = new_analysis(oss, analysis_type, settings_first=False)

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=None,
        settings=None,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)


def viewer_3d(
    oss: OpticStudioSystem,
    oncomplete: OnComplete | str = OnComplete.Release,
) -> AnalysisResult:
    """Wrapper around the OpticStudio 3D Viewer.

    Parameters
    ----------
    oss: zospy.core.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    oncomplete: OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to
        'Release'.

    Returns
    -------
    AnalysisResult
        A Viewer 3D analysis result. As the viewers do not return data, the AnalysisResult can mainly be used to
        further control the analysis when oncomplete is set to `Sustain`.
    """
    analysis_type = constants.Analysis.AnalysisIDM.Draw3D

    analysis = new_analysis(oss, analysis_type, settings_first=False)

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=None,
        settings=None,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)


def shaded_model(
    oss: OpticStudioSystem,
    oncomplete: OnComplete | str = OnComplete.Release,
) -> AnalysisResult:
    """Wrapper around the OpticStudio Shaded Model Viewer.

    Parameters
    ----------
    oss: zospy.core.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    oncomplete: OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to
        'Release'.

    Returns
    -------
    AnalysisResult
        A Shaded Model analysis result. As the viewers do not return data, the AnalysisResult can mainly be used to
        further control the analysis when oncomplete is set to `Sustain`.
    """
    analysis_type = constants.Analysis.AnalysisIDM.ShadedModel

    analysis = new_analysis(oss, analysis_type, settings_first=False)

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=None,
        settings=None,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)


def nsc_3d_layout(
    oss: OpticStudioSystem,
    oncomplete: OnComplete | str = OnComplete.Release,
) -> AnalysisResult:
    """Wrapper around the OpticStudio NSC 3D Layout viewer.

    Parameters
    ----------
    oss: zospy.core.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be non-sequential.
    oncomplete: OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to
        'Release'.

    Returns
    -------
    AnalysisResult
        A NSC 3D Layout analysis result. As the viewers do not return data, the AnalysisResult can mainly be used to
        further control the analysis when oncomplete is set to `Sustain`.
    """
    analysis_type = constants.Analysis.AnalysisIDM.NSC3DLayout

    analysis = new_analysis(oss, analysis_type, settings_first=False)

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=None,
        settings=None,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)


def nsc_shaded_model(
    oss: OpticStudioSystem,
    oncomplete: OnComplete | str = OnComplete.Release,
) -> AnalysisResult:
    """Wrapper around the OpticStudio NSC Shaded Model Viewer.

    Parameters
    ----------
    oss: zospy.core.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be non-sequential.
    oncomplete: OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to
        'Release'.

    Returns
    -------
    AnalysisResult
        A NSC Shaded Model analysis result. As the viewers do not return data, the AnalysisResult can mainly be used to
        further control the analysis when oncomplete is set to `Sustain`.
    """
    analysis_type = constants.Analysis.AnalysisIDM.NSCShadedModel

    analysis = new_analysis(oss, analysis_type, settings_first=False)

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=None,
        settings=None,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)



"""
This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from enum import Enum

__all__ = ("ISAberrationTypes", "ISFlipTypes", "ISSamplings", "ISShowAsTypes")

class ISAberrationTypes(Enum):
    None_ = 0
    Geometric = 1
    Diffraction = 2

class ISFlipTypes(Enum):
    None_ = 0
    TopBottom = 1
    LeftRight = 2
    TopBottomLeftRight = 3

class ISSamplings(Enum):
    None_ = 0
    X2 = 1
    X4 = 2
    X8 = 3
    X16 = 4
    X32 = 5
    X64 = 6

class ISShowAsTypes(Enum):
    SimulatedImage = 0
    SourceBitmap = 1
    PSFGrid = 2
